/* mem-mmap.c
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.
 *
 * Memory sub-system for systems that provide mmap.
 */

#include "ml-unixdep.h"
#include "ml-osdep.h"
#include <sys/mman.h>
#include INCLUDE_FCNTL_H
#include "ml-base.h"
#include "memory.h"

#ifdef F2A_RTS /* G.F.*/ 
#include "my_memory.h"
#endif /* G.F.*/

#if !(defined(HAS_MMAP) || defined(HAS_ANON_MMAP))
#  error expected HAS_MMAP or HAS_ANON_MMAP
#endif

/* protection mode for mmap memory */
#define PROT_ALL	(PROT_READ|PROT_WRITE|PROT_EXEC)

/* flags for mmap */
#ifdef HAS_ANON_MMAP
#  define MMAP_FLGS	(MAP_ANONYMOUS|MAP_PRIVATE)
#else
#  define MMAP_FLGS	MAP_PRIVATE
#endif

#if (defined(HOST_ALPHA32) && (defined(OPSYS_OSF1) || defined(OPSYS_DUNIX)))
  /* To insure that mmap returns a 32-bit address, we need to specify a non-zero
   * address to mmap().  The address 0x2000000 is the location of the text segment,
   * which is a good minimum value since the C stack lives just below this address
   * and mmap will find regions above it.
   */
#  define MMAP_ADDR	(caddr_t)0x2000000
#else
#  define MMAP_ADDR	0
#endif

struct mem_obj {
    Word_t	*base;	  /* the base address of the object. */
    Addr_t	sizeB;	  /* the object's size (in bytes) */
#ifdef HAS_PARTIAL_MUNMAP
#   define	mapBase		base
#   define	mapSizeB	sizeB
#else
    Word_t	*mapBase; /* base address of the mapped region containing */
			  /* the object */
    Addr_t	mapSizeB; /* the size of the mapped region containing  */
			  /* the object */
#endif
};

#ifdef OLD_RTS
extern int	errno;
#endif

#define ALLOC_MEMOBJ()		NEW_OBJ(mem_obj_t)
#define FREE_MEMOBJ(p)		FREE(p)

#include "mem-common.ins"

/* MEM_InitMemory:
 */
void MEM_InitMemory ()
{
    InitMemory();

} /* MEM_InitMemory */

/* MapMemory:
 *
 * Map a BIBOP_PAGE_SZB aligned chunk of szb bytes of virtual memory.  Return
 * the address of the mapped memory (or NIL on failure).
 */
PVT status_t MapMemory (mem_obj_t *obj, Addr_t szb)
{
    int		fd;
    Addr_t	addr, offset;

#ifdef OLD_RTS /* G.F.*/
#ifdef HAS_ANON_MMAP
    fd = -1;
#else
  /* Note: we use O_RDONLY, because some OS are configured such that /dev/zero
   * is not writable.  This works because we are using MAP_PRIVATE as the
   * mapping mode.
   */
    if ((fd = open("/dev/zero", O_RDONLY)) == -1) {
	Error ("unable to open /dev/zero, errno = %d\n", errno);
	return FAILURE;
    }
#endif
  /* we grab an extra BIBOP_PAGE_SZB bytes to give us some room for alignment */
    addr = (Addr_t) my_mmap (MMAP_ADDR, szb+BIBOP_PAGE_SZB, PROT_ALL,MMAP_FLGS, fd, 0);
    if (addr == -1) {
	Error ("unable to map %d bytes, errno = %d\n", szb, errno);
#ifndef HAS_ANON_MMAP
	close (fd); /* NOTE: this call clobbers errno */
#endif
	return FAILURE;
    }
#ifndef HAS_ANON_MMAP
    close (fd);
#endif
#else /* G.F.*/
    addr = (Addr_t)my_kmalloc(szb+BIBOP_PAGE_SZB);
    if(addr != -1)	
     	memset((void *)addr, 0, szb+BIBOP_PAGE_SZB);
    else  return FAILURE;
#endif /* G.F.*/

  /* insure BIBOP_PAGE_SZB alignment */
    offset = BIBOP_PAGE_SZB - (addr & (BIBOP_PAGE_SZB-1));
#ifdef HAS_PARTIAL_MUNMAP
    if (offset != 0) {
      /* align addr and discard unused portions of memory */
	my_munmap ((void *)addr, offset);
	addr += offset;
/*	my_munmap ((void *)(addr+szb), BIBOP_PAGE_SZB-offset); 
	I don't think I need to free this part of memory -- G.F.
*/
    }
    else {
	my_munmap ((void *)(addr+szb), BIBOP_PAGE_SZB);
    }
#else
    obj->mapBase = (Word_t *)addr;
    obj->mapSizeB = szb+BIBOP_PAGE_SZB;
    addr += offset;
#endif
    obj->base = (Word_t *)addr;
    obj->sizeB = szb;
    return SUCCESS;
} /* end of MapMemory */

/* UnmapMemory:
 *
 * Unmap a szb byte chunk of virtual memory at addr.
 */
PVT void UnmapMemory (mem_obj_t *obj)
{
    if (my_munmap((caddr_t)(obj->mapBase), obj->mapSizeB) == -1) {
#ifdef OLD_RTS 
	Die ("error my_unmapping [%#x, %#x], errno = %d\n",
	    obj->mapBase, (Addr_t)(obj->mapBase) + obj->mapSizeB, errno);
#else 
	Die ("error my_unmapping [%#x, %#x]\n",
	    obj->mapBase, (Addr_t)(obj->mapBase) + obj->mapSizeB);
#endif
    }

} /* end of UnmapMemory */


int my_munmap(void* addr, long len)
{  
#ifdef F2A_RTS /* G.F.*/
    return my_kfree(addr, len);	
#endif
#ifdef OLD_RTS
  return do_munmap(addr, len);
#endif /* G.F.*/  	
}


#ifdef F2A_RTS /* G.F.*/

int my_kfree(void * addr, long len){
  unsigned long index = (char*)addr - memory;
  block_t *next_block = &block_list;
/*
printk("my_kfree:memory=0x%lx, addr=0x%lx, index=0x%lx, len=%ld\n", 
   (unsigned long)memory, (unsigned long)addr, index, len);
*/
  if (addr==NULL) 
      return 1;

  while((next_block!=NULL)&&(next_block->index!=index))
      next_block = next_block->next;

  if (next_block == NULL){
      printk("it is not a block allocated by my_kmalloc()\n");
      return -1;
  }

  if ( len <0 )   
     len = next_block->size;

  if ( len > next_block->size) {
      printk("it is not a valid block offset\n");
      return -1;
  }


  if (len < next_block->size){
          block_t* new_block;
	  new_block = (block_t*)kmalloc(sizeof(block_t), 0x03);
          if(new_block != NULL){   
             new_block->size = len;
             new_block->index = next_block->index;
             new_block->next = next_block;
             new_block->pre =  next_block->pre;
	     if ( next_block->pre != NULL)
	         next_block->pre->next = new_block;	
             next_block->pre = new_block;
             next_block->size -= len;
             next_block->index += len;
             new_block->used =0;
	     next_block = new_block;	
	  }
	 else {
	     printk("my_kfree: no enough memory in kernel space.\n");
 	     return -1;
	 }
  }

  free_space += next_block->size;    
  if((next_block->pre != NULL)&&(!next_block->pre->used)){
        next_block->pre->size +=next_block->size;
        next_block->pre->next = next_block->next;
	if ( next_block->next != NULL)
	     next_block->next->pre = next_block->pre;
  }	

  if((next_block->next != NULL)&&(!next_block->next->used)){
        if((next_block->pre != NULL)&&(!next_block->pre->used)){
           next_block->pre->size +=next_block->next->size;
           next_block->pre->next = next_block->next->next;
           if(next_block->next->next != NULL) 
	   	next_block->next->next->pre = next_block->pre;   
           kfree(next_block->next);
           kfree(next_block);
	   return 1; 
	}
	else {
	   block_t * tmp = next_block->next;	
           next_block->size += tmp->size;
           if( tmp->next != NULL) 
	   	tmp->next->pre = next_block;   
           next_block->next = tmp->next;
	   next_block->used = 0;	
	   kfree(tmp);
           return 1;
        }
  }
  else if((next_block->pre != NULL)&&!next_block->pre->used){
        kfree(next_block);
        return 1;  
  }
  else {
	next_block->used = 0;
        return 1;
  }   
}

void* my_kmalloc(long size){
   /* always allocate memory from the tail of the first-fit free block */

   block_t *next_free_block = &block_list;
   unsigned long space = free_space; 

   if(free_space < size ){
	printk("my_kmalloc: no free space in memory\n");
	return NULL;
   }	
   /* search for a free block of size >= size, return NULL if
      unable to do so */
   while( (next_free_block != NULL)&&
	  ((next_free_block->used)||(next_free_block->size < size)) ){
	if( (!next_free_block->used)&&
            ((space -= next_free_block->size) < size) ){
	    printk(" my_kmalloc: no free space in memory\n");
            return NULL;
	}
        else next_free_block = next_free_block->next;
   }
   if (next_free_block != NULL){ /* found a block of at least size */
   	free_space -= size;
        if(next_free_block->size > size){  /* split block */
          block_t* new_block;
	  new_block = (block_t*)kmalloc(sizeof(block_t), 0x03);
          if(new_block != NULL){   
             new_block->size = size;
             new_block->index =
	       next_free_block->index + next_free_block->size - size;
             new_block->next = next_free_block->next;
             new_block->pre =  next_free_block;
	     if (new_block->next != NULL) {
	       new_block->next->pre = new_block;
	     }
             next_free_block->next = new_block;
             next_free_block->size -= size;
             new_block->used =1;
/*
printk("memory addr=(0x%lx ...0x%lx), free=%lu, size=%lu, index=%lx\n", 
      memory, memory+MEMORY_SIZE, free_space, size,new_block->index);    
*/  
	     return (void*)(memory + new_block->index);   
          }
          else {
             printk("my_kmalloc: no free memory in kernel space\n");
             return NULL;   
          }
            
        }
	else{ /* next_free_block->size == size */
          next_free_block->used = 1;
/*
printk("memory addr=(0x%lx ...0x%lx), free=%lu, size=%lu, index=%lx\n", 
      memory, memory+MEMORY_SIZE, free_space, size,next_free_block->index);    
*/  
        return (void*)(memory + next_free_block->index);   
        }    
   } 
   printk(" my_kmalloc: no free space in my memory array\n");
   return NULL;
}

#endif /* G.F. */   

