(* dev_int.sml
 * structure for device driver and interrupt handler
 *)

structure DevInt : DEV_INT  = struct
	
    type IRQ = int

    type handler = IRQ -> unit 

    type priority = int
    
    datatype dev_status = UNINSTALLED | INSTALLED of handler 	
	
    type dev_info = dev_status * bool * int * string  
		(* status * mask * priority * dev_name *)

          
    val MAX_INTERRUPT = 15 
    
    val KEYBOARD_PRIORITY:priority = 10
    val TIMER_PRIORITY:priority = 10
    val NET_PRIORITY:priority = 10
 
 	
    val devTbl = Array.array(MAX_INTERRUPT+1, (UNINSTALLED, false, ~1, "") )

    fun listHandlers() = 
	let fun listDev_loop i = 
		if i > MAX_INTERRUPT then ()
		else (
		  let val (status, _, _, name) =  Array.sub(devTbl, i)
		  in case status of 
			UNINSTALLED => () | 
			_   => print( Int.toString(i)^":"^name^"\n")
		  end;
		  listDev_loop (i+1)
	     	 )
	in listDev_loop 0
	end

    exception INVALID_INTERRUPT_ID
    exception INVALID_SHARE_INTERRUPT
	
    (* default status after installing a device handler is enabled *)
    (* enable: mask = true 	disable: mask = false *)

    fun setHandler (id, h, p, dev_name) = 
	if (id < 0) orelse ( id > MAX_INTERRUPT ) 
	then	raise INVALID_INTERRUPT_ID
	else let val (status, _, _, _) = Array.sub(devTbl, id)
	     in  case  status of 
		 UNINSTALLED =>  Array.update(devTbl, id, (h, true, p, dev_name) ) 
		|_ 	     =>  raise INVALID_SHARE_INTERRUPT;
	         status
	     end

    fun clearHandler (id ) = 
	if (id < 0) orelse ( id > MAX_INTERRUPT ) 
	then	raise INVALID_INTERRUPT_ID
	else Array.update(devTbl, id, (UNINSTALLED, false, ~1, "") )
	
    fun mask id = 
 	if (id < 0) orelse ( id > MAX_INTERRUPT ) 
	then	raise INVALID_INTERRUPT_ID
	else 
	  let val (h, _, p, name ) = Array.sub(devTbl, id)
	  in  Array.update(devTbl, id, (h, false, p, name) )
       	  end

    fun unmask id = 
        if (id < 0) orelse ( id > MAX_INTERRUPT ) 
	then	raise INVALID_INTERRUPT_ID
 	else 
	  let val (h, _, p, name ) = Array.sub(devTbl, id)
	  in  Array.update(devTbl, id, (h, true, p, name) )
       	  end	 

    exception NO_ASSOCIATED_INTERRUPT_HANDLER

    val delay_interrupts = ref false	(* don't dispatch interrupts now *)
    val delayed_interrupts = ref ([]: IRQ list)

    fun dispatch id = 
        if (id < 0) orelse ( id > MAX_INTERRUPT ) 
	then raise INVALID_INTERRUPT_ID
 	else if ! delay_interrupts then
(* list should usually be short, so this should be reasonably efficient. *)
	 delayed_interrupts := (! delayed_interrupts) @ [id]
 	else 
	  let val (h, m, p, _ ) = Array.sub(devTbl, id)
	  in  if m 
	      then(  
		   case h of 
		    	UNINSTALLED =>raise NO_ASSOCIATED_INTERRUPT_HANDLER
		       |INSTALLED hdl =>hdl id	
              )
	      else () 
	  end

    fun suspend_interrupts () =
         delay_interrupts := true

    fun resume_interrupts () =
         (delay_interrupts := false;
	  let val list = ! delayed_interrupts;
	  in delayed_interrupts := [];
	     List.app dispatch list
	  end)

   fun init() = 
	let exception Bad_Signal_In_Dev_Int
	    val s = case Signals.fromString("USR2") of
	               SOME s => s
		     | _ => raise Bad_Signal_In_Dev_Int
	    val old_h = ref( Signals.inqHandler s)
	    fun delay 0 = () | delay n = delay (n - 1)
	    fun sigHandler(signal, n, cont) = 
	     (
		let val id = ASM.getID()
	        in  if id = ~1 then 				       
 		       case !old_h of
			 Signals.HANDLER h => h(signal, n, cont) 
		         | _  => cont
		    else
		     ((dispatch id)
                      handle x => (print "exception in dispatch\n";
				   delay 100000000;
				   print ("exception " ^ General.exnName x ^" detected in interrupt handler for ID " ^ Int.toString id ^ "\n");
				   delay 100000000);
                      sigHandler(signal, n, cont))
                 end

	    )
	    val hdl = Signals.HANDLER sigHandler
	in 
	    old_h := Signals.setHandler(s, hdl)
	end

end 	
