(* structure for keyboard interrupt handler *)
structure Keyboard : KEYBOARD = struct
   val KEYBOARD_IRQ = 1

   (* keyboard controller registers *)
   val KBD_STATUS_REG   = 0x64
   val KBD_CNTL_REG     = 0x64
   val KBD_DATA_REG     = 0x60

   (* status register bits*)
   val KBD_OBF  = 0x01
   val KBD_IBF  = 0x02
   val KBD_GTO  = 0x40
   val KBD_PERR = 0x80

(* 
   change plain/shift_map 
	   [14] from 0x7f to 0x00  -- BS
	   [15] from 0x09 to 0x00  -- tab
	   [29] from 0x02 to 0x00  -- left control key
	   [42] 0x00		   -- left shift key
 	   [54] 0x00		   -- right shift key
           [55] from 0x0c to 0x00  -- * key in number-pad
   	   [56] from 0x03 to 0x00  -- left alt key
	   [58] from 0x07 to 0x00  -- caps lock
	   [59]- unchanged, all ignored
*)      
val plain_map = [
        0x00, 0x1b, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36,
        0x37, 0x38, 0x39, 0x30, 0x2d, 0x3d, 0x00, 0x00,
        0x71, 0x77, 0x65, 0x72, 0x74, 0x79, 0x75, 0x69,
        0x6f, 0x70, 0x5b, 0x5d, 0x0a, 0x00, 0x61, 0x73,
        0x64, 0x66, 0x67, 0x68, 0x6a, 0x6b, 0x6c, 0x3b,
        0x27, 0x60, 0x00, 0x5c, 0x7a, 0x78, 0x63, 0x76,
        0x62, 0x6e, 0x6d, 0x2c, 0x2e, 0x2f, 0x00, 0x00,
        0x00, 0x20, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04,
        0x05, 0x06, 0x07, 0x08, 0x09, 0x08, 0x09, 0x07,
        0x08, 0x09, 0x0b, 0x04, 0x05, 0x06, 0x0a, 0x01,
        0x02, 0x03, 0x00, 0x10, 0x06, 0x00, 0x3c, 0x0a,
        0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x0e, 0x02, 0x0d, 0x1c, 0x01, 0x05, 0x14, 0x03,
        0x18, 0x01, 0x02, 0x17, 0x00, 0x19, 0x15, 0x16,
        0x1a, 0x0c, 0x0d, 0x1b, 0x1c, 0x10, 0x11, 0x1d,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
] 

val shift_map = [
        0x00, 0x1b, 0x21, 0x40, 0x23, 0x24, 0x25, 0x5e,
        0x26, 0x2a, 0x28, 0x29, 0x5f, 0x2b, 0x00, 0x00,
        0x51, 0x57, 0x45, 0x52, 0x54, 0x59, 0x55, 0x49,
        0x4f, 0x50, 0x7b, 0x7d, 0x01, 0x00, 0x41, 0x53,
        0x44, 0x46, 0x47, 0x48, 0x4a, 0x4b, 0x4c, 0x3a,
        0x22, 0x7e, 0x00, 0x7c, 0x5a, 0x58, 0x43, 0x56,
        0x42, 0x4e, 0x4d, 0x3c, 0x3e, 0x3f, 0x00, 0x00,
        0x00, 0x20, 0x00, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
        0x0f, 0x10, 0x3b, 0x12, 0x13, 0x13, 0x03, 0x07,
        0x08, 0x09, 0x0b, 0x04, 0x05, 0x06, 0x0a, 0x01,
        0x02, 0x03, 0x00, 0x10, 0x06, 0x00, 0x3e, 0x0a,
        0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x0e, 0x02, 0x0d, 0x00, 0x01, 0x05, 0x14, 0x03,
        0x0b, 0x01, 0x02, 0x17, 0x00, 0x0a, 0x15, 0x16,
        0x1a, 0x0c, 0x0d, 0x1b, 0x1c, 0x10, 0x11, 0x1d,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
]

type 'a fifo = 'a list * 'a list
fun fifo_new() = ([], [])
fun fifo_add((fwd, back), a) =(fwd, a::back)
fun fifo_next(h::r, back) = SOME((r,back), h)  	|
    fifo_next([],[]) = NONE 	|
    fifo_next([], [single]) = SOME(([],[]), single) |
    fifo_next([], back) = 
	case rev back of 
	  h::rest => SOME((rest,[]), h) |
          _	  => NONE 

 val kbd_buf:char fifo ref = ref(fifo_new())
 
 val shift_state = ref false

 fun scancode2ascii s = 
     if(s>58) 
     then NONE
     else if (s=42) orelse (s=54)
	  then ( shift_state := true; NONE )
	  else if( !shift_state )
               then( shift_state := false; 
	             let val ascii = List.nth(shift_map,s)
		     in  if ascii <> 0 
			 then SOME ascii
			 else NONE
		     end )	 	 
               else  let val ascii = List.nth(plain_map,s)
		     in  if ascii <> 0
			 then SOME ascii
			 else NONE	
		     end
 exception INVALID_INTERRUPT

 fun interrupt (id) = 
     if id <> KEYBOARD_IRQ 
     then raise INVALID_INTERRUPT
     else
	let fun loop s = 
	    let val scancode = ASM.inb(KBD_DATA_REG) 
	    in 
	        case scancode2ascii(scancode) of
		     NONE =>()
                   | SOME ascii => 
		       kbd_buf := fifo_add(!kbd_buf, chr(ascii));
	        let val status = ASM.inb(KBD_STATUS_REG)
	        in 
                if(Word.andb(Word.fromInt status,
                             Word.fromInt KBD_OBF)<>0wx0)
                then loop status
		else()  
                end 
	    end	
	in  
	    loop (ASM.inb_p(KBD_STATUS_REG));
	(*  Scheduler.exit() *)
	    ()
	end

  fun read_char() = 
     case fifo_next(!kbd_buf) of
       SOME(buf, c) => (kbd_buf := buf; print (str(c)); SOME c)  |
       NONE => NONE 
	
  fun init() = 
      let val h = DevInt.INSTALLED interrupt
      in  DevInt.setHandler(KEYBOARD_IRQ, h, DevInt.KEYBOARD_PRIORITY, "KEYBOARD");
          ()
      end
  end 


