(* device driver for the Intel EtherExpress *)
(* translated to SML and adapted from Linux Debian version 2.0.33
   files: drivers/net/eth82586.h
          drivers/net/eexpress.c
	  linux/if_arp.h
	  linux/socket.h
	  linux/netdevice.h
   GF 2/1999
 *)

structure EExpress: EEXPRESS = struct

exception RESET_FAILURE
exception INIT_FAILURE
exception UNKNOWN_DEV_INTERRUPT
exception EEXPRESS_IN_USE
exception NO_EEXPRESS_IN_USE
exception ILLEGAL_OR_UNUSED_INTERRUPT_ID
local

(*
 *	Ethernet statistics collection data. 
 *)
 
type NetStatistics = {
	rx_packets : int ref,		(* total packets received	*)
	tx_packets : int ref,		(* total packets transmitted	*)
	rx_errors : int ref,		(* bad packets received		*)
	tx_errors : int ref,		(* packet transmit problems	*)
	rx_dropped : int ref,		(* no space in linux buffers	*)
	tx_dropped : int ref,		(* no space available in linux	*)
	multicast : int ref,		(* multicast packets received	*)
	collisions : int ref,

	(* detailed rx_errors *)
	rx_length_errors : int ref,
	rx_over_errors : int ref,		(* receiver ring buff overflow	*)
	rx_crc_errors : int ref,		(* recved pkt with crc error	*)
	rx_frame_errors : int ref,		(* recv'd frame alignment error *)
	rx_fifo_errors : int ref,		(* recv'r fifo overrun		*)
	rx_missed_errors : int ref,		(* receiver missed packet	*)

	(* detailed tx_errors *)
	tx_aborted_errors : int ref,
	tx_carrier_errors : int ref,
	tx_fifo_errors : int ref,
	tx_heartbeat_errors : int ref,
	tx_window_errors : int ref 	}


type NetLocal = {
	stats: 		NetStatistics, 
	init_time: 	int ref,
	rx_first: 	int ref,
	rx_last: 	int ref,
	tx_head: 	int ref,
	tx_reap:	int ref,
	tx_tail:	int ref,
	tx_link:	int ref,
	last_tx_restart:int ref,
	started:	int ref,
	promisc:	bool ref,
	rx_buf_start:	int ref,
	rx_buf_end:	int ref,
	num_tx_bufs:	int ref,
	num_rx_bufs:	int ref 	} 


datatype iftype = AUI | BNC | TP

type device = {
          name: string, 
	  base_addr: int, (* device IO address *)
	  irq: DevInt.IRQ,  (* device IRQ number *)
	  if_port: iftype,
	  start: bool ref,
	  interrupt: bool ref, (* interrupt arrived *)
	  tbusy: bool ref, 
	  trans_start: int ref,
	  last_rx: int ref,
	  flags: int ref, 	
	  family : int, 
	  priv: NetLocal, (* private *)
	  broadcast: Word8Vector.vector, (* broadcast[MAX_ADDR_LEN ] *)
 	  dev_addr: Word8Vector.vector,  (* dev_addr[MAX_ADDR_LEN *)
	  addr_len: int,
	  mtu: int,
	  hard_header_len: int,
	  hardware_type: int,  
	  InComingBuffers: (Word8Array.array list) ref    	
}  

fun int_eq (x: int, y) = x = y

(* map irq number to device structure *)
structure Dev_Map = Single_Store (structure V = Fox_Basis.V
				  type key = int
				  type value = device
				  val eq = int_eq
				  val hash = Word.fromInt)


(* following constant value came from eth82586.h *)

(*
 * EtherExpress card register addresses

 * as offsets from the base IO region (dev->base_addr) 
 *)
val DATAPORT  = 0x0000
val WRITE_PTR = 0x0002
val READ_PTR = 0x0004
val SIGNAL_CA = 0x0006
val SET_IRQ = 0x0007
val SM_PTR = 0x0008
val MEM_Ctrl = 0x000b
val MEM_Page_Ctrl = 0x000c
val Config = 0x000d
val EEPROM_Ctrl = 0x000e
val ID_PORT = 0x000f

(*
 * offset to shadowed memory, 0 <= x <= 31. We don't use this yet,
 * but may in the future. Is shadow memory access any faster than
 * dataport access?
 *)
fun SM_ADDR(x) = Word.toInt(0wx4000 + Word.<<(Word.andb(x, 0wx10), 0w10) + Word.andb(x, 0wxf))

(* Always mirrors eexp-memory at 0x0008-0x000f *)
val SCB_STATUS = 0xc008
val SCB_CMD = 0xc00a
val SCB_CBL = 0xc00c
val SCB_RFA = 0xc00e

(*
 * card register defines
 *)

(* SET_IRQ *)
val SIRQ_en = 0wx08
val SIRQ_dis  = 0wx00

(* Config *)
fun set_loopback(ioaddr) =
     ASM.outb(Word.toInt(Word.orb(Word.fromInt(ASM.inb(ioaddr+Config)), 0wx02 )),
	      (ioaddr+Config))

fun clear_loopback(ioaddr) =
     ASM.outb(Word.toInt(Word.andb(Word.fromInt(ASM.inb(ioaddr+Config)), 0wxfd)),
	      (ioaddr+Config)) 



(* EEPROM_Ctrl *)
val EC_Clk = 0wx01 	(* EEPROM shift clock *)
val EC_CS  = 0wx02	(* EEPROM Chip Select *)
val EC_Wr  = 0wx04	(* EEPROM chip data in *)
val EC_Rd  = 0wx08	(* EEPROM chip data out *)
val ASIC_RST = 0wx40	
val i586_RST = 0wx80

val EEPROM_DELAY_TIME = 0

fun eeprom_delay() = 
    let fun eeprom_delay_loop i = if( i > 0 ) then eeprom_delay_loop(i-1) else ()	
    in  eeprom_delay_loop EEPROM_DELAY_TIME
    end
(*
 * i82586 Memory Configuration
 *)

(* (System Configuration Pointer) System start up block, read after 586_RST *)
val SCP_START = 0xfff6


(* Intermediate System Configuration Pointer *)
val ISCP_START = 0x0000
(* System Command Block *)
val SCB_START = 0x0008

(*
 * Start of buffer region. If we have 64k memory, eexp_hw_probe() may raise
 * NUM_TX_BUFS. RX_BUF_END is set to the end of memory, and all space between
 * the transmit buffer region and end of memory used for as many receive buffers
 * as we can fit. See eexp_hw_[(rx)(tx)]init().
 *)
val TX_BUF_START = 0x0100
val TX_BUF_SIZE = Word.toInt(Word.andb( Word.fromInt((24+ EthConstant.ETH_FRAME_LEN+31)),
			    	   	Word.notb(0wx1f)))
val RX_BUF_SIZE = Word.toInt(Word.andb( Word.fromInt((32+ EthConstant.ETH_FRAME_LEN+31)),
				       Word.notb(0wx1f)))

(*
 * SCB defines
 *)

(* these functions take the SCB status word and test the relevant status bit *)
fun SCB_complete(s) = Word.andb(Word.fromInt(s), 0wx8000) <> 0w0
fun SCB_rxdframe(s) = Word.andb(Word.fromInt(s), 0wx4000) <> 0w0
fun SCB_CUdead(s) = Word.andb(Word.fromInt(s), 0wx2000) <> 0w0
fun SCB_RUdead(s) = Word.andb(Word.fromInt(s), 0wx1000) <> 0w0
fun SCB_ack(s) = Word.andb(Word.fromInt(s), 0wxf000) 

(* Command unit status: 0=idle, 1=suspended, 2=active *)
fun SCB_CUstat(s) = Word.toInt(Word.>>(Word.andb(Word.fromInt(s), 0wx0300), 0w8))

(* Receive unit status: 0=idle, 1=suspended, 2=out of resources, 4=ready *)
fun SCB_RUstat(s) = Word.toInt(Word.>>(Word.andb(Word.fromInt(s), 0wx0070), 0w4))

(* SCB commands *)
val SCB_CUnop     = 0wx0000
val SCB_CUstart   = 0wx0100
val SCB_CUresume  = 0wx0200
val SCB_CUsuspend = 0wx0300
val SCB_CUabort   = 0wx0400
val SCB_resetchip = 0wx0080
val SCB_RUnop     = 0wx0000
val SCB_RUstart   = 0wx0010
val SCB_RUresume  = 0wx0020
val SCB_RUsuspend = 0wx0030
val SCB_RUabort   = 0wx0040

(* 
 *
 * Command block defines
 *)
fun Stat_Done(s)=   Word.andb(Word.fromInt s, 0wx8000) <> 0w0
fun Stat_Busy(s)=   Word.andb(Word.fromInt s, 0wx4000) <> 0w0
fun Stat_OK(s)  =   Word.andb(Word.fromInt s, 0wx2000) <> 0w0
fun Stat_Abort(s)=  if Word.andb(Word.fromInt s, 0wx1000) = 0w0 then 0 else 1
fun Stat_STFail(s)  =  Word.andb(Word.fromInt s, 0wx0800) <> 0w0
fun Stat_TNoCar(s)= if Word.andb(Word.fromInt s, 0wx0400) = 0w0 then 0 else 1
fun Stat_TNoCTS(s)= if Word.andb(Word.fromInt s, 0wx0200) = 0w0 then 0 else 1
fun Stat_TNoDMA(s)= if Word.andb(Word.fromInt s, 0wx0100) = 0w0 then 0 else 1
fun Stat_TDefer(s)= Word.andb(Word.fromInt s, 0wx0080) <> 0w0
fun Stat_TColl(s)=  Word.andb(Word.fromInt s, 0wx0040) <> 0w0
fun Stat_TXColl(s)= Word.andb(Word.fromInt s, 0wx0020) <> 0w0
fun Stat_NoColl(s)= Word.toInt(Word.andb(Word.fromInt s, 0wx000f))

(* Cmd_END will end AFTER the command if this is the first
 * command block after an SCB_CUstart, but BEFORE the command
 * for all subsequent commands. Best strategy is to place
 * Cmd_INT on the last command in the sequence, followed by a
 * dummy Cmd_Nop with Cmd_END after this.
 *)
val Cmd_END  =   0x8000
val Cmd_SUS  =   0x4000
val Cmd_INT  =   0x2000

val Cmd_Nop     = 0x0000
val Cmd_SetAddr = 0x0001
val Cmd_Config  = 0x0002
val Cmd_MCast   = 0x0003
val Cmd_Xmit    = 0x0004
val Cmd_TDR     = 0x0005
val Cmd_Dump    = 0x0006
val Cmd_Diag    = 0x0007


(*
 * Frame Descriptor (Receive block) defines
 *)

fun FD_Done(s)=  Word.andb(Word.fromInt s, 0wx8000) <> 0w0
fun FD_Busy(s)=  Word.andb(Word.fromInt s, 0wx4000) <> 0w0
fun FD_OK(s)  =  Word.andb(Word.fromInt s, 0wx2000) <> 0w0

fun FD_CRC(s)  = if (Word.andb(Word.fromInt s, 0wx0800) = 0w0) then 0 else 1
fun FD_Align(s)= if (Word.andb(Word.fromInt s, 0wx0400) = 0w0) then 0 else 1
fun FD_Resrc(s)= if (Word.andb(Word.fromInt s, 0wx0200) = 0w0) then 0 else 1
fun FD_DMA(s)  = if (Word.andb(Word.fromInt s, 0wx0100) = 0w0) then 0 else 1
fun FD_Short(s)= if (Word.andb(Word.fromInt s, 0wx0080) = 0w0) then 0 else 1
fun FD_NoEOF(s)= if (Word.andb(Word.fromInt s, 0wx0040) = 0w0) then 0 else 1


(* following constant value came from linux/drivers/net/eexpress.c *)

val EEXPRESS_IO_EXTENT = 16

val CONF_LINK = 0x0020
val CONF_HW_ADDR = 0x0038

val irqrmap = [	0wx0, 0wx0, 0wx1, 0wx2, 0wx3, 0wx4, 0wx0, 0wx0,
		0wx0, 0wx1, 0wx5, 0wx6, 0wx0, 0wx0, 0wx0, 0wx0  ]

val start_code:Word8.word list = [
	0wx00, 0wx00,                 (* SCP: set bus to 16 bits *)
	0wx00, 0wx00, 0wx00, 0wx00,          (* junk *)
	0wx00, 0wx00, 0wx00, 0wx00,          (* address of ISCP (lo,hi) *)

	0wx01, 0wx00,                 (* ISCP: busy - cleared after reset *)
	0wx08, 0wx00, 0wx00, 0wx00,0wx00, 0wx00,   (* offset,address (lo,hi) of SCB *)

	0wx00, 0wx00, 0wx00, 0wx00, (* SCB: status, commands *)
	0wx00, 0wx00, 0wx00, 0wx00, (* links to first command block, first receive descriptor *)
	0wx00, 0wx00, 0wx00, 0wx00, (* CRC error, alignment error counts *)
	0wx00, 0wx00, 0wx00, 0wx00, (* out of resources, overrun error counts *)

	0wx00, 0wx00, 0wx00, 0wx00,          (* pad *)
	0wx00, 0wx00, 0wx00, 0wx00,

	0wx00, 0wx00, 0wx02, 0wx00, (* Cmd_Config, startup configure sequence, at 0wx0020 *)
	0wx32, 0wx00,                 (* link to next command *)
	0wx0c, 0wx08,          (* 12 bytes follow : fifo threshold=8 *)
	0wx40, 0wx2e,  (* don't rx bad frames : SRDY/ARDY => ext. sync. : 
			* preamble len=8 take addresses from data buffers : 6 bytes/address 
			*)
	0wx00, 0wx60,  (* default backoff method & priority : interframe spacing = 0x60 *)
	0wx00, 0wxf2,   (* slot time=0x200 : max collision retry = 0xf *)
	0wx00, 0wx00,   (* no HDLC : normal CRC : 
			 * enable broadcast : disable promiscuous/multicast modes 
			 *)
	0wx3c, 0wx00,                 (* minimum frame length = 60 octets) *)

	0wx00, 0wx00, 0wx01, 0wx20,   (* Cmd_INT | Cmd_SetAddr *)
	0wx3e, 0wx00,                 (* link to next command *)
	0wx00, 0wx00, 0wx00, 0wx00, 0wx00, 0wx00,   (* hardware address placed here, 0x0038 *)
	0wx00, 0wx00, 0wx00, 0wx80,   (* Cmd_END | Cmd_Nop *)
				(* end of configure sequence *)
	0wx3e, 0wx00,

	0wx00, 0wx00
]

(* 
 * following constant value came from Donald Becker's Diagnostic 
 *  program for Intel EtherExpress 
 *)

val EE_WRITE_CMD = 0wx140  (* 5 << 6 *)
val EE_READ_CMD	 = 0wx180  (* 6 << 6 *)
val EE_ERASE_CMD = 0wx1c0  (* 7 << 6 *)

fun get_vector_addr( p:Word8Vector.vector ref): Word32.word = Unsafe.cast ( p);
fun get_array_addr(p): Word32.word = Unsafe.cast ( p)
fun get_array_ref_addr(p:Word8Array.array ref): Word32.word = Unsafe.cast( p)

(*
fun debug_loop(step) =  (* 20MHz Pentium it takes 0.1 us (micro second) per loop *) 
    let fun loop i = if i=0 then () else loop(i-1)
    in loop(step)
    end 
*)

fun print_in(buf, len) = 
    let 
        fun p i  = 
            if i = len 
	    then print("\n")
	    else(print (Word8.toString(Word8Array.sub(buf, i)) ^ "."); 
		 p(i+1) )
    in print "in : ";
       p 0
    end

fun print_out(buf, len) = 
    let 
        fun p i  = 
            if i = len 
	    then print("\n")
	    else(print (Word8.toString(Word8Vector.sub(buf, i)) ^ "."); 
		 p(i+1) )
    in print "out: ";
       p 0
    end

datatype TR = Device of device

fun newBuf (dev: device, size) = 
    let val buf = Word8Array.array(size, 0wx0)
    in #InComingBuffers dev := (!(#InComingBuffers dev))@[buf];
       buf
    end

fun delBuf (dev: device) = 
     case !(#InComingBuffers dev) of
            [] => Word8Array.array(0, 0wx0)
          | h::t => ( #InComingBuffers dev := t; 
		      h )


fun eexpress_hw_readeeprom(ioaddr, location) = 
    let val cmd = Word.orb(EE_READ_CMD, Word.andb(Word.fromInt(location),0wx7f))
	val ee_addr = ioaddr + EEPROM_Ctrl
	val wval = Word.orb(EC_CS, i586_RST)

	fun write_eeprom(v) = (	
   	    ASM.outb(v, ee_addr);
	    ASM.outb(Word.toInt(Word.orb(Word.fromInt(v), EC_Clk)), ee_addr);
            eeprom_delay();
            ASM.outb(v, ee_addr);
            eeprom_delay() ) 
	
 	fun read_eeprom(r, i) = (
	    ASM.outb(Word.toInt(Word.orb(wval, EC_Clk)), ee_addr);
	    eeprom_delay();
	    if (Word.andb(Word.fromInt(ASM.inb ee_addr), EC_Rd)<>0w0)
	    then ( ASM.outb(Word.toInt wval, ee_addr);
	    	   eeprom_delay();
		   Word.orb(r, i) )
	    else ( ASM.outb(Word.toInt wval, ee_addr);
	    	   eeprom_delay();
		   r ) )

	fun terminate_eeprom() = (
	    ASM.outb(Word.toInt(Word.orb(i586_RST, EC_Clk)), ee_addr);
	    eeprom_delay();
	    ASM.outb(Word.toInt i586_RST, ee_addr);
	    eeprom_delay()  )

	fun loop1(w, i) = 
	    if (i = 0w0 ) 
	    then ()
	    else if (Word.andb(cmd,i) <> 0w0 )
	         then let val new_w = Word.orb(w, EC_Wr)
		      in  write_eeprom(Word.toInt(new_w));
		          loop1(new_w, Word.>>(i,0w1))
		      end		
		 else let val new_w = Word.andb(w, Word.notb(EC_Wr))
		      in  write_eeprom(Word.toInt(new_w));
			  loop1(new_w, Word.>>(i, 0w1))
		      end 	

	fun loop2(r, i) = 
	    if ( i = 0w0 )
	    then r
	    else loop2(read_eeprom(r, i), Word.>>(i, 0w1))
			    	
    in 
	ASM.outb(Word.toInt wval, ee_addr);
	loop1(wval, 0wx100);
	ASM.outb(Word.toInt wval, ee_addr);
	loop2(0w0, 0wx8000) before terminate_eeprom()
    end  
		

fun eexpress_hw_txinit (dev:device) = 
    let val ioaddr = #base_addr dev
	val lp = #priv dev
	val num_tx_bufs = !(#num_tx_bufs(lp))
	val old_wp = ASM.inw(ioaddr+WRITE_PTR)
	
	fun loop(tx_block, curtbuf) = 
	    if ( curtbuf = num_tx_bufs )
	    then tx_block
	    else(
		  ASM.outw(tx_block, ioaddr+WRITE_PTR);
		  ASM.outw(0x0, ioaddr);
		  ASM.outw(Word.toInt(Word.orb(Word.fromInt(Cmd_INT),
					       Word.fromInt(Cmd_Xmit))), ioaddr);
		  ASM.outw(tx_block+0x08, ioaddr);
		  ASM.outw(tx_block+0x0e, ioaddr);
		  ASM.outw(0x0, ioaddr);
		  ASM.outw(0x0, ioaddr);
		  ASM.outw(tx_block+0x08, ioaddr);
		  ASM.outw(0x8000, ioaddr);
		  ASM.outw(~1, ioaddr);
		  ASM.outw(tx_block+0x16, ioaddr);
		  ASM.outw(0x0, ioaddr);
		  loop(tx_block+TX_BUF_SIZE, curtbuf + 1) )	

        val tx_blocks = loop(TX_BUF_START, 0);
   in   #tx_head(lp) := TX_BUF_START;
	#tx_reap(lp) := TX_BUF_START;
	#tx_tail(lp) := tx_blocks - TX_BUF_SIZE;
	#tx_link(lp) := tx_blocks - TX_BUF_SIZE + 0x08;
	#rx_buf_start(lp) := tx_blocks;
	ASM.outw(old_wp, ioaddr + WRITE_PTR)
   end	

fun eexpress_hw_rxinit (dev:device) = 
    let val ioaddr = #base_addr dev
	val lp = #priv dev
	val rx_words = Word8Vector.fromList(
		     [  0wxed, 0wxfe, 0wx0d, 0wxf0, 0wx01, 0wxf0, 0wx05, 
			0wx05, 0wx24, 0wx24, 0wx65, 0wx65, 0wxaf, 0wxde ])
	val old_wp = ASM.inw(ioaddr+WRITE_PTR)

        fun loop(num_rx_bufs, rx_block) = 
	     (ASM.outw(rx_block, ioaddr+WRITE_PTR);
	      ASM.outw(0x0, ioaddr);
	      ASM.outw(0x0, ioaddr);
	      ASM.outw(rx_block+RX_BUF_SIZE, ioaddr);
	      ASM.outw(rx_block+0x16, ioaddr);
	      ASM.outsw(ioaddr, get_vector_addr(ref rx_words), 
			(Word8Vector.length(rx_words)) div 2);
	      ASM.outw(0x8000, ioaddr);
	      ASM.outw(~1, ioaddr);
	      ASM.outw(rx_block+0x20, ioaddr);
	      ASM.outw(0x0, ioaddr);
	      ASM.outw(Word.toInt(Word.orb(0wx8000, Word.fromInt(RX_BUF_SIZE-0x20))),
		       ioaddr);
	      if (rx_block > (!(#rx_buf_end lp))) then (num_rx_bufs, rx_block) (* return *)
	      else loop(num_rx_bufs+1, rx_block+RX_BUF_SIZE) )

    	val (num_rx_buf, new_rx_block) = loop(0, !(#rx_buf_start lp)) 
   in   #num_rx_bufs(lp) := num_rx_buf;
	#rx_first(lp) := !(#rx_buf_start lp);
	#rx_last(lp) := new_rx_block;
	ASM.outw(!(#rx_last lp) + 4, ioaddr+WRITE_PTR);
   	ASM.outw(!(#rx_first lp), ioaddr);
	ASM.outw(old_wp, ioaddr+WRITE_PTR)
    end	


fun eexpress_hw_ASICrst(ioaddr, irq) =
    let val rst = Word.toInt(Word.orb(ASIC_RST, i586_RST))
	fun loop1(wrval, succount, boguscount) = ( 
	    ASM.outw(0, ioaddr + WRITE_PTR);
	    ASM.outw(wrval, ioaddr);
	    ASM.outw(0, ioaddr + READ_PTR);
    	    if ( wrval = ASM.inw(ioaddr))
	    then loop2(wrval+1, succount+1, boguscount)
	    else if boguscount = 1
		 then ( print "having problem reseting EExpress ASIC, continuing ...\n";
			ASM.outb(rst, ioaddr+EEPROM_Ctrl);
			loop2(0x1, 0, 500) )
		 else loop2(wrval + 1, 0, boguscount-1) )
       and
	   loop2(wrval, succount, boguscount) = 
	   if succount >= 20
	   then ()
	   else if wrval = 0xffff
		then loop1(0x1, succount, boguscount)
		else loop1(wrval, succount, boguscount)
	
    in 
	ASM.outb(Word.toInt(Word.orb(SIRQ_dis, List.nth(irqrmap, irq))), 
		 ioaddr + SET_IRQ);
	ASM.outb(rst, ioaddr + EEPROM_Ctrl);
	loop2(0x1, 0, 500);
	ASM.outb(Word.toInt(i586_RST), ioaddr + EEPROM_Ctrl);
	0 (* priv->started := 0 *)
    end


fun eexpress_hw_init586 (dev:device) =  
    let val lp = #priv dev
	val ioaddr = #base_addr dev

	fun loop4(iboguscount, ifailcount) = 
	    if ASM.inw(ioaddr + SCB_STATUS) = 0
	    then if iboguscount = 1
		 then if ifailcount > 0

		      then ( print("i82586 init timed out, state =" ^
				   Int.toString(ASM.inw(ioaddr + SCB_STATUS))^" cmd= "^
				   Int.toString(ASM.inw(ioaddr + SCB_CMD))^
				   "try again\n");
			     ASM.outw(CONF_LINK,ioaddr+SCB_CBL);
			     ASM.outw(0,ioaddr+SCB_STATUS);
			     ASM.outw(Word.toInt(Word.orb(0wxf000, SCB_CUstart)),
				      ioaddr+SCB_CMD);
			     ASM.outb(0,ioaddr+SIGNAL_CA);
			     loop4(100, ifailcount - 1) )
		      else ( print("init i82586 fail\n");
			     raise INIT_FAILURE	)
		 else loop4(iboguscount - 1, ifailcount)
	    else () 	

	fun loop3(rboguscount, rfailcount) = (
	    ASM.outw(0, ioaddr+READ_PTR);	
	    if ASM.inw(ioaddr) = 0
	    then ()
	    else if (rboguscount = 1)
		 then (
			print("i82586 reset timed out, signal CA again.\n");
			ASM.outw(0,ioaddr+SCB_CMD);
			ASM.outb(0,ioaddr+SIGNAL_CA);	 
			if rfailcount = 1
			then ( print("i82586 is not responding, reset fail.\n"); 
			       raise RESET_FAILURE )
			else loop3(100, rfailcount-1) )
		else loop3( rboguscount-1, rfailcount)  
	)

        fun loop2 i = 
	    if i = 0w16 
	    then ()
	    else ( ASM.outw_p(0, ioaddr+SM_ADDR(Word.<<(i, 0w1)));
		   loop2 (i+0w1) ) 

	fun loop1 ofs = 
	    if ofs >= !(#rx_buf_end lp)
	    then ()
	    else ( ASM.outw_p(ofs, ioaddr+SM_PTR);
		   loop2 0w0;
		   loop1 (ofs+32) )
    in  
	#started(lp) := 0;
	set_loopback(ioaddr);
	ASM.outb(Word.toInt(Word.orb(SIRQ_dis, List.nth(irqrmap,#irq(dev)))),
		 ioaddr+SET_IRQ);
	ASM.outb_p(Word.toInt(i586_RST),ioaddr+EEPROM_Ctrl);
	loop1 0;
	ASM.outw_p(!(#rx_buf_end lp), ioaddr+WRITE_PTR);
        #promisc(lp) := Word.andb(Word.fromInt(!(#flags dev)), NetDevConstant.IFF_PROMISC) <> 0w0; 
	let val buf = if !(#promisc lp)
		       then let val i = Word8.orb(List.nth(start_code, 28*2+1), 0wx1)
			    in 
			       Word8Vector.fromList(List.take(start_code, 28*2+1) @ 
							       [i] @ List.drop(start_code, 29*2))
			     end
		      else
		       let val i = Word8.andb(List.nth(start_code, 28*2+1), 0wxfe)
		       in 
			  Word8Vector.fromList(List.take(start_code, 28*2+1) @ 
					       [i] @ List.drop(start_code, 29*2))
		       end
	in ASM.outsw(ioaddr, get_vector_addr(ref buf), (Word8Vector.length(buf)+1) div 2 )
	end;
	ASM.outw(CONF_HW_ADDR,ioaddr+WRITE_PTR);
	ASM.outsw(ioaddr,get_vector_addr(ref(#dev_addr dev)), 
		  (Word8Vector.length(#dev_addr dev)+1) div 2);
	eexpress_hw_txinit(dev);
	eexpress_hw_rxinit(dev);
	ASM.outw(0,ioaddr+WRITE_PTR);
	ASM.outw(1,ioaddr);
	ASM.outb(0,ioaddr+EEPROM_Ctrl);
	ASM.outw(0,ioaddr+SCB_CMD);
	ASM.outb(0,ioaddr+SIGNAL_CA);		
	loop3(50, 5);
	ASM.outw(CONF_LINK,ioaddr+SCB_CBL);
	ASM.outw(0,ioaddr+SCB_STATUS);
	ASM.outw(Word.toInt(Word.orb(0wxf000, SCB_CUstart)),ioaddr+SCB_CMD);
	ASM.outb(0,ioaddr+SIGNAL_CA);
	loop4(50, 5);
	ASM.outb(Word.toInt(Word.orb(SIRQ_en, List.nth(irqrmap, #irq(dev)))),
		 ioaddr+SET_IRQ);
	clear_loopback(ioaddr);
	#init_time(lp) := Timer.gettime()
    end	


fun eexpress_hw_txrestart(dev:device) = print("this should never happen!!!\n")

fun eexpress_hw_tx(buf, dev:device) =
    let val ioaddr = #base_addr dev
	val len = Word8Array.length buf
	val lp = #priv dev
        val old_wp = ASM.inw(ioaddr + WRITE_PTR )
	val tx_head = !(#tx_head lp)
	val tx_tail = !(#tx_tail lp)
    in  ASM.outw(tx_head, ioaddr + WRITE_PTR);
	ASM.outw(0x0, ioaddr);
	ASM.outw(Word.toInt(Word.orb(Word.fromInt(Cmd_INT), 
                                     Word.fromInt(Cmd_Xmit))), ioaddr);
	ASM.outw(tx_head+0x8, ioaddr);
	ASM.outw(tx_head+0xe, ioaddr);
	ASM.outw(0x0, ioaddr);
	ASM.outw(0x0, ioaddr);
	ASM.outw(tx_head+0x8, ioaddr);
	ASM.outw(Word.toInt(Word.orb(0wx8000, Word.fromInt len)), ioaddr);
	ASM.outw(~1, ioaddr);
	ASM.outw(tx_head+0x16, ioaddr);
	ASM.outw(0, ioaddr);
(*
	ASM.outsw(ioaddr, get_vector_addr(ref buf), (len+1) div 2);
*)
	ASM.outsw(ioaddr, get_array_ref_addr(ref buf), (len+1) div 2);
	ASM.outw(tx_tail+0xc, ioaddr+WRITE_PTR);
	ASM.outw(tx_head, ioaddr);
	#trans_start dev := Timer.gettime();
	#tx_tail lp := tx_head;
	if tx_head = (TX_BUF_START+(!(#num_tx_bufs lp)-1)*TX_BUF_SIZE)
	then #tx_head lp := TX_BUF_START
	else #tx_head lp := tx_head + TX_BUF_SIZE;
	#tbusy dev := (!(#tx_head lp) = !(#tx_reap lp));
	ASM.outw(old_wp, ioaddr + WRITE_PTR);
(*
	print_out(buf, len)
*)
	()
    end


fun eexpress_hw_lasttxstat (dev:device) = 
    let val lp = #priv dev
        val ioaddr = #base_addr dev
        val old_rp = ASM.inw(ioaddr+READ_PTR)
        val old_wp = ASM.inw(ioaddr+WRITE_PTR)
	fun loop1() = 
  	    let val status = ASM.inw(ioaddr)
	    in  if not (Stat_Done(status))
		then ( 
		      (* 
		        print "lasttxstat: transmission not complete\n";
		       *)
		        #tx_link(lp) := !(#tx_reap lp); 
			status )
		else let val s = #stats lp
		     in  #last_tx_restart(lp) := 0;
			 #collisions(s):= !(#collisions s) + Stat_NoColl(status);
			 if not (Stat_OK(status))
			 then ( print "hw_lasttxstat: errors in transmission detected\n";
			        #tx_aborted_errors(s) := !(#tx_aborted_errors(s)) + 
			 			        Stat_Abort(status);
		 	        #tx_carrier_errors(s) := !(#tx_carrier_errors(s)) +
				 Word.toInt(Word.orb(Word.fromInt(Stat_TNoCar(status)),
					 	     Word.fromInt(Stat_TNoCTS(status))));
				#tx_fifo_errors(s) := !(#tx_fifo_errors(s)) + 
						      Stat_TNoDMA(status) )
			 else #tx_packets(s) := !(#tx_packets s) +1;
			 if !(#tx_reap lp) = ((!(#num_tx_bufs lp)-1)*TX_BUF_SIZE+TX_BUF_START)
			 then #tx_reap(lp) := TX_BUF_START
			 else #tx_reap(lp) := !(#tx_reap(lp)) + TX_BUF_SIZE;
			 #tbusy(dev) := false; 
			 loop(false, status) 
		     end 
	    end
    and				
	 loop (isFirst, s) = 
	    if isFirst 
	    then ( ASM.outw(!(#tx_reap lp), ioaddr+READ_PTR);
		   loop1() )
	    else if (!(#tx_reap lp) = !(#tx_head lp))
	    	 then ( #tx_link(lp) := !(#tx_tail lp) + 0x08;
			s )
	    	 else ( ASM.outw(!(#tx_reap lp), ioaddr+READ_PTR);
		        loop1() )
			
    in  if !(#tbusy dev) andalso (!(#tx_head lp) = !(#tx_reap lp))
	then (print "lasttxstat: immediate exit\n"; 0)
	else let val s = loop(true, 0);
  	     in ASM.outw(old_rp,ioaddr+READ_PTR);
	        ASM.outw(old_wp,ioaddr+WRITE_PTR);
		s
	     end
    end  	   	    	


fun eexpress_hw_rx (dev:device) =
    let val ioaddr = #base_addr dev
	val old_wp = ASM.inw(ioaddr + WRITE_PTR)
	val old_rp = ASM.inw(ioaddr + READ_PTR )
	val hard_header_len = #hard_header_len dev
	val lp = #priv dev
	val stats = #stats lp
	val original_boguscount = !(#num_rx_bufs(lp))
	
	fun loop(rx_block, boguscount) = 
	    (ASM.outw(rx_block, ioaddr+READ_PTR);
	     if boguscount <= 0 then ()
	     else
	      let 
	          val status = ASM.inw(ioaddr)
		  val rfd_cmd = ASM.inw(ioaddr)
		  val rx_next = ASM.inw(ioaddr)
		  val pbuf    = ASM.inw(ioaddr)

		  fun func_1 pkt_len = 
		       if (rfd_cmd <> 0) orelse ( pbuf <>(rx_block+0x16)) 
			    orelse ( Word.andb(Word.fromInt(pkt_len), 0wxc000) <> 0wxc000)
		        then 
			 (print "corrupt frame received\n";
			  loop(rx_block, boguscount))
		        else if not (FD_OK(status))
			     then(
				  print "error frame received\n";
			          #rx_errors(stats) := !(#rx_errors(stats)) + 1;
			          #rx_crc_errors(stats) := !(#rx_crc_errors(stats))
			 				    + FD_CRC(status);
			          #rx_frame_errors(stats) := !(#rx_frame_errors(stats))
							    + FD_Align(status);
			          #rx_fifo_errors(stats):= !(#rx_fifo_errors(stats))
							    + FD_Resrc(status);
			          #rx_over_errors(stats) := !(#rx_over_errors(stats))
							    + FD_DMA(status);
			          #rx_length_errors(stats) := !(#rx_length_errors(stats))
							    + FD_Short(status); 
			         ASM.outw(rx_block, ioaddr+WRITE_PTR);
			         ASM.outw(0x0, ioaddr);
			         ASM.outw(0x0, ioaddr);
			         loop(rx_next, boguscount-1) )
			    else let 
			             val pkt_len = Word.toInt(Word.andb(Word.fromInt(pkt_len),
			   				                0wx3fff)) 
				     val buf = newBuf(dev, pkt_len) 
				     val addr = get_array_addr(ref buf)
			         in 
				    ASM.outw(pbuf+10, ioaddr+READ_PTR);
				    ASM.insw(ioaddr, addr, (pkt_len+1) div 2);
				(*  
				    print_in(buf, pkt_len);
				 *) 
				    #rx_packets(stats) := !(#rx_packets(stats)) + 1;
 			     	    ASM.outw(rx_block, ioaddr+WRITE_PTR);
			     	    ASM.outw(0x0, ioaddr);
			     	    ASM.outw(0x0, ioaddr);
				    loop(rx_next, boguscount-1) 
			         end

	        in if FD_Done(status)
		    then(ASM.outw(pbuf, ioaddr+READ_PTR);
			 func_1( ASM.inw ioaddr ) )  
		    else( (* ASM.outw(rx_next, ioaddr+READ_PTR); *)
			 loop(rx_next, boguscount-1) )
	        end)
    in loop(!(#rx_first(lp)), !(#num_rx_bufs(lp)));
       ASM.outw(old_rp, ioaddr+READ_PTR);
       ASM.outw(old_wp, ioaddr+WRITE_PTR)
    end   		

fun init_stats() = {
	rx_packets = ref 0,		(* total packets received	*)
	tx_packets = ref 0,		(* total packets transmitted	*)
	rx_errors = ref 0,		(* bad packets received		*)
	tx_errors = ref 0,		(* packet transmit problems	*)
	rx_dropped = ref 0,		(* no space in linux buffers	*)
	tx_dropped = ref 0,		(* no space available in linux	*)
	multicast = ref 0,		(* multicast packets received	*)
	collisions = ref 0,

	(* detailed rx_errors *)
	rx_length_errors = ref 0,
	rx_over_errors = ref 0,		(* receiver ring buff overflow	*)
	rx_crc_errors = ref 0,		(* recved pkt with crc error	*)
	rx_frame_errors = ref 0,		(* recv'd frame alignment error *)
	rx_fifo_errors = ref 0,		(* recv'r fifo overrun		*)
	rx_missed_errors = ref 0,		(* receiver missed packet	*)

	(* detailed tx_errors *)
	tx_aborted_errors = ref 0,
	tx_carrier_errors = ref 0,
	tx_fifo_errors = ref 0,
	tx_heartbeat_errors = ref 0,
	tx_window_errors = ref 0 	}


fun init_priv(ioaddr, irq) = { 
    	         stats = init_stats(),
		 init_time = ref 0,
		 rx_first = ref 0,
		 rx_last = ref 0,
		 tx_head = ref 0,
		 tx_reap = ref 0,
		 tx_tail = ref 0,
		 tx_link = ref 0,
		 last_tx_restart = ref 0,
		 started = ref (eexpress_hw_ASICrst(ioaddr, irq)),
		 promisc = ref false,
		 rx_buf_start = ref 0,
		 rx_buf_end = ref 0,
		 num_tx_bufs = ref 0,
		 num_rx_bufs = ref 0	} 	  

fun eexpress_hw_probe(name:string, ioaddr) = 
    let val hw_addr = [ eexpress_hw_readeeprom(ioaddr, 2),
			eexpress_hw_readeeprom(ioaddr, 3),
			eexpress_hw_readeeprom(ioaddr, 4) ]
	fun get_buf_info(lp:NetLocal) = 
	    let val i586mso = 0x023e
		val old_wp = ASM.inw(ioaddr+WRITE_PTR)
		val old_rp = ASM.inw(ioaddr+READ_PTR) before
			     ASM.outw(0x8000+i586mso,ioaddr+READ_PTR)
		val old_a1 = ASM.inw(ioaddr) before 
			     ASM.outw(i586mso,ioaddr+READ_PTR)
		val old_a0 = ASM.inw(ioaddr) before (
			     ASM.outw(i586mso,ioaddr+WRITE_PTR);
	 		     ASM.outw(0x55aa,ioaddr);
			     ASM.outw(i586mso,ioaddr+READ_PTR) )
		val a0_0 =   ASM.inw(ioaddr) before (
	  		     ASM.outw(0x8000+i586mso,ioaddr+WRITE_PTR);
			     ASM.outw(0x5a5a,ioaddr);
			     ASM.outw(0x8000+i586mso,ioaddr+READ_PTR) )
		val a1_0 =   ASM.inw(ioaddr) before
	 		     ASM.outw(i586mso,ioaddr+READ_PTR)
		val a0_1 =   ASM.inw(ioaddr) before (
			     ASM.outw(i586mso,ioaddr+WRITE_PTR);
			     ASM.outw(0x1234,ioaddr);
			     ASM.outw(0x8000+i586mso,ioaddr+READ_PTR) )
		val a1_1 =   ASM.inw(ioaddr)
		
	   in   
		if (a0_0 <> a0_1 )  orelse (a1_0 <> a1_1) orelse 
		   (a1_0 <> 0x5a5a) orelse (a0_0 <> 0x55aa )
		then (#num_tx_bufs(lp) := 4;
		      #rx_buf_end(lp) := 0x7ff6 ) 
		else (#num_tx_bufs(lp) := 8;
		      #rx_buf_start(lp) := TX_BUF_START + 8 * TX_BUF_SIZE;
		      #rx_buf_end(lp) := 0xfff6 );  
		ASM.outw(0x8000+i586mso,ioaddr+WRITE_PTR);
	  	ASM.outw(old_a1,ioaddr);
		ASM.outw(i586mso,ioaddr+WRITE_PTR);
		ASM.outw(old_a0,ioaddr);
		ASM.outw(old_wp,ioaddr+WRITE_PTR);
		ASM.outw(old_rp,ioaddr+READ_PTR); 
		lp
	   end

        fun get_ether_addr(addr_in) = 
	    let 
	        fun loop (i, l) = 
                    if i = 3  then l 
                    else 
		     loop(i+1,
			  Word8.fromInt
			  (Word.toInt(Word.>>
				      (Word.andb(List.nth(addr_in,i),
						 0wxff00), 
				       0w8))) ::
			  (Word8.fromInt(Word.toInt(List.nth(addr_in, i)))
			       :: l ))
	    
	   in Word8Vector.fromList(loop(0, []:Word8.word list))
	   end
	
	fun set_irq() = 
	    let val irqmap = [ 0, 9, 3, 4, 5, 10, 11, 0 ]
		val setupval = eexpress_hw_readeeprom(ioaddr, 0)
		val v5 = eexpress_hw_readeeprom(ioaddr, 0)
		val irq = List.nth(irqmap, Word.toInt(Word.>>(setupval, 0wx0d)))
		val if_port = 	if ( Word.andb(setupval, 0wx1000) = 0wx0 )
				then AUI 
				else if ( Word.andb(v5, 0wx1) = 0wx0 )
				     then BNC
				     else TP
	    in  ASM.outb(Word.toInt(Word.orb(SIRQ_dis, List.nth(irqrmap, irq))),ioaddr+SET_IRQ);
		ASM.outb(0,ioaddr+SET_IRQ);
		(irq, if_port)
	    end

      fun get_broadcast_addr() = 
	  let fun loop(l, i) = 
	      	  if i = 0 
	      	  then l
	      	  else loop(l@[0wxff], i-1)
	  in  Word8Vector.fromList(loop([]:Word8.word list, EthConstant.ETH_ALEN))
	  end
		
    in( 
	if (List.nth(hw_addr, 2) <> 0wx00aa ) orelse 
	   (Word.andb(List.nth(hw_addr, 1), 0wxff00) <> 0wx0 )
	then NONE  (* NET_ENODEV *)
	else  let val (irq, if_port) = set_irq()
		  val d:device = { 
			    name = name,
			    base_addr = ioaddr,
			    dev_addr = get_ether_addr(hw_addr),
	          	    irq = irq,
			    if_port = if_port,
			    start = ref false,
			    interrupt = ref false,
			    tbusy = ref false,
			    trans_start = ref 0,
			    last_rx = ref 0,		
			    flags = ref (Word.toInt(Word.orb(NetDevConstant.IFF_BROADCAST, 
							    NetDevConstant.IFF_MULTICAST))),
			    family = NetDevConstant.AF_INET,
			    priv = get_buf_info(init_priv(ioaddr, irq)),	
			    broadcast = get_broadcast_addr(),
			    addr_len = EthConstant.ETH_ALEN,
			    mtu = 1500,	
			    hard_header_len = EthConstant.ETH_HLEN,
			    hardware_type = NetDevConstant.ARPHRD_ETHER,
			    InComingBuffers = ref []
			}		
	      in  
		  SOME d 	
	      end
    )
    end 	


fun eexpress_probe(name, base_addr) = 
    let val ports = [ 0x300, 0x270, 0x320, 0x340, 0 ]
	val ioaddr = base_addr

	fun get_magic_eeprom (i, sum, port) = 
	    if i = 4
	    then sum 
	    else let val t = Word.fromInt(ASM.inb(port+ID_PORT))
		 in 
		    get_magic_eeprom(i+1, Word.orb(sum, 
                                                    Word.<<(Word.>>(t, 0w4),
				      			Word.<<(Word.andb(t, 0wx3), 0w2))), 
				      port)
		 end

	fun check_eeprom i = 
	    let val port = List.nth(ports, i)
	    in ( 
		if port = 0
		then (  print("probe: ENODEV\n");
			NONE (* NET_ENODEV *) )
		else if(get_magic_eeprom(0, 0w0, port) = 0wxbaba )
		     then case eexpress_hw_probe(name, port) of
				SOME v =>SOME v  (* NET_DEVICE *)
			      | NONE  => check_eeprom (i+1)
		     else check_eeprom (i+1) )
	     end
				     
    in  if (Word.andb(Word.fromInt(ioaddr), 0wxfe00)<>0wx0)
	then eexpress_hw_probe(name, base_addr)
	else if (ioaddr <>0 ) 
	     then ( print("probe: ENXIO\n"); NONE (* NET_ENXIO *) )
	     else check_eeprom 0
    end	




fun eexpress_xmit(buf:Word8Array.array, dev:device) = 
    let 
        val ioaddr = #base_addr dev
	val irq = #irq dev
	val lp = #priv dev
	fun resetCU v = (
	    ASM.outw(v,ioaddr+SCB_CBL);
	    ASM.outw(0,ioaddr+SCB_STATUS);
	    ASM.outw(Word.toInt(SCB_CUstart),ioaddr+SCB_CMD);
	    ASM.outb(0,ioaddr+SIGNAL_CA) )
	fun loop (boguscount) =
	    if SCB_complete(ASM.inw(ioaddr+SCB_STATUS))
	    then ()
	    else if boguscount = 1
		 then ( print("reset timed out, retrying ...\n");
			resetCU (!(#tx_link lp));
			loop 200 )
		 else loop( boguscount-1 )	
    in  ASM.outb(Word.toInt(Word.orb(SIRQ_dis, List.nth(irqrmap, irq))),ioaddr+SET_IRQ);
	if !(#tbusy dev) 
	then if !(#started lp) = 0 
	     then if (Timer.gettime() - !(#trans_start dev)) > 5 
		  then if !(#tx_link lp) = !(#last_tx_restart lp)
		       then (
			      print("Retransmit timed out, resetting ... \n");
			      eexpress_hw_txinit (dev);
			      #last_tx_restart(lp) := 0;
			      resetCU(!(#tx_link lp));		
			      loop 200;
			      #tbusy dev := false 
			)
			else if SCB_CUdead(ASM.inw(ioaddr+SCB_STATUS))
			     then( let val _  = eexpress_hw_lasttxstat(dev)
				   in print("Transmit timed out, CU not active, restarting...\n");
				      eexpress_hw_txrestart dev
				   end )
			     else let val ret_zero = eexpress_hw_lasttxstat(dev)
			          in if (!(#tbusy dev) andalso (ret_zero=0) )
				     then( print("CU wedged, resetting ...\n");
				           eexpress_hw_init586 dev;
				           #tbusy dev := false)
				     else ()
				  end     
		 else () 
    	    else if (Timer.gettime()-(!(#init_time lp))) > 10
	         then ( print("i82586 startup timed out, resetting ...\n");
		        eexpress_hw_init586 dev;
		    	#tbusy dev := false
	     	 )
	     	 else ()
	else ();	   		
	if Word8Array.length(buf) = 0
	then if SCB_CUdead(ASM.inw(ioaddr+SCB_STATUS))
	     then( print("CU has died, restarting...\n");
		   eexpress_hw_lasttxstat dev;
		   #tx_errors(#stats lp) := !(#tx_errors(#stats lp))+1;	
		   eexpress_hw_txrestart dev
	     )
	     else(print "why this happen ?\n";  
		  eexpress_hw_lasttxstat dev; 
		  ()
		  (*   ASM.cli() *)
	     )    	
	else if !(#tbusy(dev))
	     then ( print("transmitter access conflict, dropping the packet\n");
		    #tx_dropped(#stats lp) := !(#tx_dropped(#stats lp))+1 )
	     else(  #tbusy(dev) := true;  
		    ASM.outb(Word.toInt(Word.orb(SIRQ_dis, List.nth(irqrmap, irq))),ioaddr+SET_IRQ);
		    eexpress_hw_tx(buf,dev));
		    (* print "transmited a packet\n" );*) 
	ASM.outb(Word.toInt(Word.orb(SIRQ_en, List.nth(irqrmap, irq))),ioaddr+SET_IRQ)
    end	



fun eexpress_irq irq = 
    let val enter_time = Timer.gettime()
        val dev:device = Dev_Map.find irq
	val ioaddr = #base_addr dev
	val call_receiver = ref false 
	fun func_1 () = 
	    let val old_rp = ASM.inw(ioaddr + READ_PTR)
		val old_wp = ASM.inw(ioaddr + WRITE_PTR)
		val status = ref (ASM.inw(ioaddr + SCB_STATUS))
		val ack_cmd = ref (SCB_ack(!status))
		val lp = #priv dev
		fun loop s = 
		    if SCB_CUstat(s)=2 
		    then loop(ASM.inw_p(ioaddr+SCB_STATUS))
		    else s
	    in		
		if (!(#started(lp)) = 0) andalso SCB_complete(!status)
		then ( 	
			print("lower half: SCBcomplete event received\n");
			status := loop(!status);
			print("lower half: CU went non-active\n");
			#started(lp) := 1;
			ASM.outw_p(!(#tx_link(lp)), ioaddr+SCB_CBL);
			ASM.outw_p(!(#rx_buf_start(lp)), ioaddr+SCB_RFA);
			ack_cmd := Word.orb(!ack_cmd, 
					   Word.orb(SCB_CUstart, SCB_RUstart)) 
		)
		else if !(#started(lp)) <> 0
		then ( (*  
		      print("lower half: updating packet transmission statistics\n"); *)
			eexpress_hw_lasttxstat dev; ())
		else ();
		if SCB_rxdframe(!status)
		then ( 	(* print "lower half: got a packet\n"; *)
			eexpress_hw_rx dev;
			call_receiver := true)
		else ();
		if (Word.andb(Word.fromInt(!(#started lp)), 0wx2)<>0w0) 
		    andalso (SCB_RUstat(!status)<>4)
		then( 
		      print("lower half: restart RU ...\n");	
		      #rx_errors(#stats lp) := !(#rx_errors(#stats lp))+1;
		      eexpress_hw_rxinit(dev); 	
	       	      ASM.outw(!(#rx_buf_start lp), ioaddr+SCB_RFA);
		      ack_cmd := Word.andb(!ack_cmd, SCB_RUstart)
		)
		else if (!(#started lp)=1) andalso (SCB_RUstat(!status)=4)
		     then #started lp :=Word.toInt(Word.orb(Word.fromInt(!(#started lp)), 0w2))
		     else ();	
		ASM.outw(Word.toInt(!ack_cmd),ioaddr+SCB_CMD);
		ASM.outb(0,ioaddr+SIGNAL_CA);
		ASM.outw(old_rp,ioaddr+READ_PTR);
		ASM.outw(old_wp,ioaddr+WRITE_PTR)
	    end
    in 
	#interrupt dev := true;
	ASM.outb(Word.toInt(Word.orb(SIRQ_dis, List.nth(irqrmap, irq))), ioaddr+SET_IRQ); 
	func_1();
	ASM.outb(Word.toInt(Word.orb(SIRQ_en, List.nth(irqrmap,irq))), ioaddr+SET_IRQ);
	#interrupt dev := false;
(*	Fox_Basis.Scheduler.handler_data_ready irq *)
	!call_receiver
(*
;
	print("time for handling eexpress interrupt="^Int.toString((Timer.gettime()-enter_time)*10)^"ms\n") *)
    end



fun install_int_handler (dev: device) = 
    let (*  val h = DevInt.INSTALLED eexpress_irq *)
	val irq = #irq dev
	val ioaddr = #base_addr dev
	val name = #name dev
    in  if (irq=0) orelse (List.nth(irqrmap, irq)=0wx0)
	then raise ILLEGAL_OR_UNUSED_INTERRUPT_ID
	else( 
              case Dev_Map.look(irq) of 
		 NONE =>  
		  ( (* DevInt.setHandler(irq, h, DevInt.NET_PRIORITY, "EEXPRESS"); *)
		   Dev_Map.add(irq, dev);
		   if ASM.request_irq(irq, 0, name)= 1
		   then ( print "eexpress in use!\n";
			  DevInt.clearHandler(irq);
			  Dev_Map.remove (irq);
			  raise EEXPRESS_IN_USE)
		   else ())
		|_  => raise EEXPRESS_IN_USE)
    end

fun eexpress_open(dev:device) = (
    install_int_handler(dev);  
    ASM.request_region(#base_addr(dev), EEXPRESS_IO_EXTENT, "eexpress");
    #tbusy(dev) := false;
    #interrupt(dev) := false;		
    eexpress_hw_init586(dev);
    #start(dev) := true
)


fun eexpress_close (dev:device) = 
    let val irq = #irq dev
	val ioaddr = #base_addr dev
	val name = #name dev 
    in  #tbusy(dev) := true;
	#start(dev) := false;
     	ASM.outb(Word.toInt(Word.orb(SIRQ_dis, List.nth(irqrmap, irq))),
                 ioaddr+SET_IRQ);
	ASM.outw(Word.toInt(Word.orb(SCB_CUsuspend,SCB_RUsuspend)), ioaddr+SCB_CMD);
	ASM.outb(0,ioaddr+SIGNAL_CA);
	DevInt.clearHandler irq;
	Dev_Map.remove irq;
	ASM.free_irq irq;
	ASM.outb(Word.toInt(i586_RST),ioaddr+EEPROM_Ctrl);
	ASM.release_region(ioaddr, EEXPRESS_IO_EXTENT)
    end


fun eexpress_stats (dev:device) = #stats(#priv(dev))

fun eexpress_receive (dev: device) = 
    let val buf = delBuf dev
        val len = Word8Array.length buf
    in  (buf, len)
    end		

fun eexpress_set_multicast (dev:device) = 
    if (Word.andb(Word.fromInt(!(#flags(dev))), NetDevConstant.IFF_PROMISC)<>0w0 ) 
	<> !(#promisc(#priv(dev)))
    then eexpress_hw_init586 dev
    else ()

fun eexpress_etheraddress (dev: device) = #dev_addr dev

in
(* following function are added for integration with foxnet   03/08/99 -- GF  *)
type T = device
type irq = int

fun probe(name, addr) = 
    case eexpress_probe(name, addr) of 
       NONE => raise UNKNOWN_DEV_INTERRUPT
      |SOME dev => dev

val dev_open = eexpress_open

fun get_ethernet_address (dev: T) = 
    let val addr = #dev_addr dev
        val len = Word8Vector.length addr
	val buf = Word8Array.array(len, 0wx0)
    in  Word8Array.copyVec{di = 0, dst = buf, len = SOME len, si = 0, src = addr};
        buf
    end

val dev_close = eexpress_close

fun send(dev, buf)= eexpress_xmit(buf, dev)

fun receive(dev, dst, i) = 
    let val src = delBuf dev
        val len = Word8Array.length src
    in  Word8Array.copy {di = i, dst = dst, len = SOME len, si = 0, src = src};
	len
    end		    

fun irqToDev irq =
     Dev_Map.look irq

fun devToIrq (dev: T) = #irq dev

fun handle_irq incoming_packet_handler irq =
     if eexpress_irq irq then incoming_packet_handler irq else ()


end 
end 






