(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

Kerberos Key Distribution Center interface.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Krb_KDC
	2.	fun kdc_address
	3.	fun rpc

		iii.	RCS Log
	
$Log: krbkdc.fun,v $
Revision 1.3  1994/05/25  09:52:09  robby
Ken's revisions following a code review

Revision 1.2  1993/12/21  13:52:15  cline
Added high_priority_filter to BuildUdp functor.

Revision 1.1  1993/12/14  21:11:01  cline
Initial revision



		1.	functor Krb_KDC
 *)

functor Krb_KDC(structure B        : FOX_BASIS
                structure Protocol : KRB_PROTOCOL
                  sharing type Protocol.realm = string
                structure Realm    : KRB_REALM
                structure Convert  : CONVERT
                val udp_header_size : int     (* 42?    *)
                val ip_over_eth     : ubyte2  (* 2ux800 *)
                val udp_over_ip     : ubyte1  (* 1u17   *)
                val udp_local_port  : ubyte2  (* "random" port number *)
                val max_retries     : int
                val sleep_time      : int
                val do_prints       : bool): KRB_KDC =
struct
  exception No_Reply

  type realm = Protocol.realm

  val do_if_debug = if B.Debug.include_prints andalso do_prints then
                     (fn (f, x) => if ! B.Debug.do_prints then f x else ())
		    else (fn (f, x) => ())

  structure Scheduler = Coroutine (structure Debug = B.Debug
				   structure V = B.V
				   val do_prints = do_prints)


  structure Event_Queue = Event_Queue(structure Scheduler = Scheduler
                                      structure Debug     = Debug
                                      structure V         = Fox_Basis.V
                                      val do_prints       = do_prints)

  structure Udp_Stack = BuildUdp(structure Scheduler      = Scheduler
                                 structure Event_Queue    = Event_Queue
                                 val ip_over_eth          = ip_over_eth
                                 val udp_over_ip          = udp_over_ip
                                 val do_prints            = do_prints
                                 val high_priority_filter = true
                                 val protocol_numbers     = [ip_over_eth])

  structure Ip      = Udp_Stack.Ip
  structure Resolve = Udp_Stack.Resolve
  structure Udp     = Udp_Stack.Udp

  val (local_ip, local_interface, gateway_ip) = Resolve.local_info ()

(*
	2.	fun kdc_address
 *)
  fun kdc_address realm =
    let
      val kdc_ip  = Resolve.name_map
                      (nth (Realm.kerberos_servers realm, 0))
      val ip_addr = Ip.Address {ip=kdc_ip, proto=udp_over_ip}
    in
      Udp.Address { remote_peer = ip_addr,
                    local_port  = udp_local_port,
                    remote_port = Protocol.port}
    end

(*
	3.	fun rpc

	Send data to KDC and return the reply from the KDC.
 *)

  fun rpc (realm, data) =
    let
      val outgoing_packet =
        Send_Packet.create
          (Convert.flat_vector2bytearray (Locative.locget data),
           udp_header_size)

      val result = (ref NONE):ByteArray.bytearray option ref

      fun handler conn incoming =
        (
          do_if_debug (B.V.print,
                       "(krbkdc.fun) handler: Got one (size=" ^
                       (Integer.makestring
                          (ByteArray.length
                             (Receive_Packet.read incoming))) ^
                       ")!\n");
          if !result <> NONE
            then B.V.print "send_to_kdc: unexpected packet!\n"
            else ();
          result := SOME (Receive_Packet.read incoming);
          Udp.close conn;
          Udp.finalize ();
          do_if_debug (B.V.print,
                       (Format.makestring_bytearray
                         (Receive_Packet.read incoming)) ^
                        "\n")
        )

      fun resend_process count =
        (case !result of
           SOME reply =>
             locative (Convert.bytearray2flat_vector reply)
         | _ =>
           (
             Scheduler.sleep ("(krbkdc.fun)resend-process", sleep_time);
             if count >= max_retries
               then raise No_Reply
               else resend_process (count+1)
           )
        ) handle x => (Udp.finalize (); raise x)
    in
      do_if_debug (B.V.print,
                   "(krbkdc.fun):sending packet to KDC:" ^
                   (Format.makestring_bytearray
                     (Convert.flat_vector2bytearray (Locative.locget data))) ^
                   "\n");

      Udp.initialize (); (* finalize is done by handler or resend_process *)

      Ip.control (Ip.Set_Interface_Address { interface = local_interface,
                                             ip_number = local_ip});

      Udp.send (Udp.active_open (kdc_address realm, handler))
               outgoing_packet;

      resend_process 0
    end

end (* functor *)
