(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

Miscellaneous functions for Kerberos


		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Krb_Util
	2.	fun read_password
	3.	fun crypt
	4.	fun parse_ticket
	5.	fun get_tgticket

		iii.	RCS Log
	
$Log: krbutil.fun,v $
Revision 1.4  1994/05/24  14:33:35  robby
 Ken's restructuring following a code review.

Revision 1.3  1993/12/15  15:22:36  cline
Removed #s from changelog.

Revision 1.2  1993/12/14  21:10:24  cline
Reorganization of code.  This version actually runs (for kinit).

Revision 1.1  1993/11/17  18:16:40  cline
Initial revision


		1.	functor Krb_Util
 *)

functor Krb_Util (structure B        : FOX_BASIS
                  structure Protocol : KRB_PROTOCOL
                    sharing type Protocol.name     = string
                        and type Protocol.instance = string
                        and type Protocol.realm    = string
                  structure KDC      : KRB_KDC
                    sharing type KDC.realm = Protocol.realm
                  structure DES          : DES
                  structure Convert      : CONVERT
		  structure Swap_Bytes   : SWAP_BYTES
                  structure Forge        : FORGE
                  val do_prints          : bool) : KRB_UTIL =
struct

  exception Bad_Protocol_Version       (*server protocol version <> ours*)
  exception Bad_Reply                  (*invalid packet returned from server*)
  exception Bad_Password               (*cannot decrypt cyphertext*)
  exception Kdc_Error of ubyte4*string (*server error reply*)

  type principal = Protocol.principal
  type local_principal = Protocol.local_principal
  type credential = Protocol.credential

  type protocol_message_header =
    {| version       : ubyte1,
       auth_msg_type : ubyte1,
       principal     :  {| name     : ubyte1 flat_vector,
                           instance : ubyte1 flat_vector,
                           realm    : ubyte1 flat_vector |},
       time_sec      : Unaligned.Byte4u.ubytes |}

  structure F = Forge

  fun debug_print s =
    if do_prints then B.V.print s else ()

  val !& = Locative.locget

  fun flatten_string s =
    Convert.string2flat_vector (s^"\000")

  fun build_message {message_type,
                     principal as (Protocol.Principal
                                     {name, instance, realm}),
                     data} =
    let
      val _ = debug_print "(krbutil.fun)enter build_message\n"
      val B.V.Misc.TIME {sec, ...} = B.V.Misc.gettimeofday ()
      val message =
        {| version       = Protocol.protocol_version,
           auth_msg_type = Protocol.auth_message_type2byte message_type,
           principal     =  {| name     = flatten_string name,
                               instance = flatten_string instance,
                               realm    = flatten_string realm |},
           time_sec      = Unaligned.Byte4u.from_int sec,
           data          = data |} 
    in
      debug_print "(krbutil.fun)leave build_message\n";
      locative (message (*: Protocol.message*))
    end

(*
		2.	fun read_password

Turn off echo, read password, turn on echo, and convert to DES key.
 *)
  fun read_password () =
    let
      fun remove_trailing_newline s =
        substring (s, 0, (size s) - 1)
    in
      System.Unsafe.CInterface.system "/bin/stty -echo";
      B.V.print "Password: ";
      DES.cblock_to_ubyte
        (DES.string_to_key (remove_trailing_newline (input_line std_in)))
      before
        (B.V.print "\n"; System.Unsafe.CInterface.system "/bin/stty echo")
    end

(*
		3.	fun crypt

	Generic interface to DES encrypt/decrypt functions
 *)
  local
    fun crypt encrypt (text, get_key) =
      let
        val _ = debug_print "(krbutil.fun)crypt: getting key\n"
        val key = DES.ubyte_to_cblock (get_key ())
        val schedule = DES.key_schedule key
        val _ = debug_print "(krbutil.fun)crypt: calling pcbc_encrypt\n"
      in
        DES.pcbc_encrypt (text, schedule, key, encrypt)
      end
  in
    val encrypt = crypt true
    val decrypt = crypt false
  end

(*
		4.	fun parse_ticket
 *)
  fun parse_ticket (reply, get_key) =
    let
      val _ = debug_print "(krbutil.fun)parse_ticket: parsing...\n"

      val &{| version       = r_version,
              auth_msg_type = r_auth_msg_type,
              principal     = r_principal,
              time          = r_time,
              body          = r_body |} =
        (F.forge (F.Forge_Tuple [F.Forge_Byte1,
                                 F.Forge_Byte1,
                                 F.Forge_Tuple [F.Forge_CString,
                                                F.Forge_CString,
                                                F.Forge_CString],
                                 F.Forge_Byte4u,
                                 F.Forge_Vector desc_byte1]) reply):
            {| version        : ubyte1,
               auth_msg_type  : ubyte1,
               principal : {| name     : ubyte1 flat_vector,
                              instance : ubyte1 flat_vector,
                              realm    : ubyte1 flat_vector |},
               time           : Unaligned.Byte4u.ubytes,
               body           : ubyte1 flat_vector |} locative

      (* CHECK PACKET VERSION OF THE RETURNED PACKET *)
      val _ = debug_print "(krbutil.fun)parse_ticket: checking packet...\n"
      val _ = if !& r_version <> Protocol.protocol_version then
              raise Bad_Protocol_Version else ()

      (* CHECK BYTE ORDER *)
      val _ = debug_print "(krbutil.fun)parse_ticket: checking byte order...\n"
      val {swap_ubyte2, swap_ubyte4, ...} =
        if (Byte1.&& (!& r_auth_msg_type, 1u1)) = Protocol.byte_order
          then Swap_Bytes.no_swap
          else Swap_Bytes.do_swap

      (* CHECK REPLY TYPE *)
      val _ = debug_print "(krbutil.fun)get_tkt_from_kdc: checking reply type...\n"
      fun err_reply_exception message_body =
        (* take bytes containing 4byte error code and string and raise
           an exception *)
        let
          val _ = debug_print "(krbutil.fun)get_tkt_from_kdc: err_reply_exception ("
          val &{| code, text |}
              = F.forge
                  (F.Forge_Tuple [F.Forge_Byte4u, F.Forge_Vector desc_byte1])
                      message_body
              : {| code : Unaligned.Byte4u.ubytes,
                   text : ubyte1 flat_vector |} locative
          val arg1 = swap_ubyte4 (Unaligned.Byte4u.align (!& code))
          val arg2 = Convert.flat_vector2string (!& text)
        in
          debug_print (arg2 ^ ")\n");
          raise Kdc_Error (arg1, arg2)
        end

      val _ = case Protocol.byte2auth_message_type
                     (Locative.locget r_auth_msg_type) of
                Protocol.Kdc_Reply => ()
              | Protocol.Err_Reply => err_reply_exception r_body
              | _ => raise Bad_Reply

      (* EXTRACT INFORMATION FROM RETURN PACKET *)
      val _ = debug_print "(krbutil.fun)parse_ticket: extracting information...\n"
      val _ = debug_print ("(krbutil.fun)parse_ticket: from `" ^
                             (Convert.flat_vector2string
                                (Locative.locget r_body)) ^ "'\n")
      val &{| ntickets, exp_date, kvno, cipher_len_b2u, rest0 |}
          = F.forge (F.Forge_Tuple [F.Forge_Byte1,
                                    F.Forge_Byte4u,
                                    F.Forge_Byte1,
                                    F.Forge_Byte2u,
                                    F.Forge_Vector desc_byte1]) r_body
          : {| ntickets       : ubyte1, (*Always zero for protocol version 4*)
               exp_date       : Unaligned.Byte4u.ubytes,
               kvno           : ubyte1,
               cipher_len_b2u : Unaligned.Byte2u.ubytes,
               rest0          : ubyte1 flat_vector |} locative
      val cipher_len = Byte2.to_int
                         (swap_ubyte2
                            (Unaligned.Byte2u.align (!& cipher_len_b2u)))
      val _ = debug_print  ("(krbutil.fun)parse_ticket: cipher_len = "^
                            (Integer.makestring cipher_len) ^ "\n");
      val cipher = F.forge (F.Forge_Vector_Length (desc_byte1, cipher_len))
                           rest0

      val _ = debug_print "(krbutil.fun)parse_ticket: forged cipher\n"

      val decrypted = decrypt (cipher, get_key)
                      handle _ => raise Bad_Password

      val _ = debug_print  ("(krbutil.fun)parse_ticket: decrypted to:  "^
                            (Format.makestring_bytearray
                              (Convert.flat_vector2bytearray
                                (Locative.locget decrypted))) ^
                            "\n")

      val &{| session_key, service, lifetime, kvno, ticket_length, rest1 |}
          = (F.forge (F.Forge_Tuple [F.Forge_Vector_Length (desc_byte1, 8),
                                     F.Forge_Tuple [F.Forge_CString,
                                                    F.Forge_CString,
                                                    F.Forge_CString],
                                     F.Forge_Byte1,
                                     F.Forge_Byte1,
                                     F.Forge_Byte1,
                                     F.Forge_Vector desc_byte1]) decrypted
             handle _ => raise Bad_Password)
          : {| session_key   : ubyte1 flat_vector, (* length 8 *)
               service       : {| name     : ubyte1 flat_vector,
                                  instance : ubyte1 flat_vector,
                                  realm    : ubyte1 flat_vector |},
               lifetime      : ubyte1,
               kvno          : ubyte1,
               ticket_length : ubyte1,
               rest1         : ubyte1 flat_vector |} locative
      val rest1_len = case Locative.locdesc rest1 of
                        desc_vector (l, _) => l
                      | _ => let exception impossible in raise impossible end
      val &{| ticket, time_sec, dummy |}
          = F.forge (F.Forge_Tuple
                       [F.Forge_Vector_Length
                         (desc_byte1, Byte1.to_int
                                        (Locative.locget ticket_length)),
                        F.Forge_Byte4u,
                        F.Forge_Vector (desc_byte1)]) rest1
          : {| ticket   : ubyte1 flat_vector,
               time_sec : Unaligned.Byte4u.ubytes,
               dummy    : ubyte1 flat_vector |} locative

      fun unflatten loc =
        let
          val s = Convert.flat_vector2string (!& loc)
        in substring (s, 0, size(s) - 1) end

      val &{| name=s_name, instance=s_instance, realm=s_realm |} = service
    in
      Protocol.Credential
        { session_key = Protocol.Key (!& session_key),
          service     = Protocol.Principal
                          { name     = unflatten s_name,
                            instance = unflatten s_instance,
                            realm    = unflatten s_realm },
          lifetime    = !& lifetime,
          kvno        = !& kvno,
          ticket      = Protocol.Ticket (!& ticket),
          issue_date  = swap_ubyte4 (Unaligned.Byte4u.align (!& time_sec)) }
    end

(*
		5.	fun get_tgticket

Get a ticket granting ticket from the key distribution center.
 *)
  fun get_tgticket {principal,
                    service as (Protocol.Local_Principal
                                  {name=sname, instance=sinstance}),
                    life} =
    let
      (* BUILD REQUEST PACKET *)
      val _ = debug_print "(krbutil.fun)get_tkt_from_kdc: building request...\n"
      val request_packet =
        let
          val body = locative {| lifetime   = life,
                                 s_name     = flatten_string sname,
                                 s_instance = flatten_string sinstance |}
          val data = Locative.locget (Convert.flatten body)
        in
          debug_print "(krbutil.fun)get_tkt_from_kdc: building message\n";
          build_message
            { message_type = Protocol.Kdc_Request,
              principal     = principal,
              data = data }
        end

      (* SEND THE REQUEST AND RECEIVE THE RETURN PACKET *)
      val _ = debug_print "(krbutil.fun)get_tkt_from_kdc: sending...\n"
      val reply = case principal of
                    Protocol.Principal {realm, ...} =>
                      KDC.rpc (realm, (Convert.flatten request_packet))
    in
      parse_ticket (reply, fn () => read_password  ())
    end (* get_tgticket *)

  fun request { principal as Protocol.Principal {name, instance, realm},
                service   as Protocol.Local_Principal
                               {name = sname, instance = sinstance},
                srealm,         (* server's realm *)
                checksum:ubyte4,(* optional checksum *)
                lifetime:ubyte4,
                tg_credential as
                  Protocol.Credential
                    { session_key = tg_session_key,
                      kvno        = kvno_from_ticket,
                      ticket      = tg_ticket,
                      service     = _,
                      lifetime    = _,
                      issue_date  = _}
              } =
    let
      val B.V.Misc.TIME {sec, usec} = B.V.Misc.gettimeofday ()
      fun get_tg_session_key () = case tg_session_key of
                                    Protocol.Key key => locative key;
      fun flat_vector_length v =
          Byte1.from_int (Flat_Vector.flat_length v)
      val authenticator =
        Locative.locget
          (encrypt
            ( Convert.flatten
                ( locative
                    {| principal  = {| name = flatten_string name,
                                       instance = flatten_string instance,
                                       realm    = flatten_string realm |},
                       checksum   = Unaligned.Byte4u.unalign checksum,
                       local_time = {| time_5ms = 1u0 (*Byte1.from_int usec*),
                                       time_sec = Unaligned.Byte4u.from_int sec
                                     |}
                     |}),
              get_tg_session_key))
      val _ = debug_print "(krbutil.fun)request: building appl_request\n"
      val Protocol.Ticket tgticket = tg_ticket
      val appl_request =
        locative
          {| version       = Protocol.protocol_version,
             auth_msg_type = Protocol.auth_message_type2byte
                               Protocol.Appl_Request,
	     kvno          = kvno_from_ticket,	(*server's key version*)
	     realm         = flatten_string realm,
  	     ticket_length = flat_vector_length tgticket,
             authenticator_length
                           = flat_vector_length authenticator,
             ticket        = tgticket,
             authenticator = authenticator,
             time_ws       = Unaligned.Byte4u.from_int 0,
             lifetime      = Byte1.from_int (Byte4.to_int lifetime),
             service       = {| name     = flatten_string sname,
                                instance = flatten_string sinstance |} |}

     val _ = debug_print "(krbutil.fun)request: converting appl_request\n"
     val flat_request = Convert.flatten appl_request

     val _ = debug_print "(krbutil.fun)request: sending RPC to KDC\n"
     val reply = KDC.rpc (realm, (Convert.flatten appl_request))

    in
      parse_ticket (reply, get_tg_session_key)
    end (* request *)
end
