(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Nick Haines  (Nick.Haines@cs.cmu.edu)
	Edoardo Biagioni <Edoardo.Biagioni@cs.cmu.edu>
        Ken Cline    (Ken.Cline@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	Functions for manipulating the kerberos ticket file.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Krb_Tkt_File
	2.	types shared with Protocol
	3.	fun is_whitespace
	4.	fun input_string
	5.	fun input_byte4
	6.	fun getenv
	7.	fun ticketfile_name
	8.	fun input_ticketfile
	9.	fun clear_ticketfile
	10.	fun append_ticketfile


		iii.	RCS Log
	
$Log: tktfile.fun,v $
Revision 1.3  1994/08/03  19:52:44  esb
got rid of '#'s in RCS log

Revision 1.2  1994/07/15  18:43:04  robby
*** empty log message ***

Revision 1.1  1994/01/14  17:42:18  cline
Initial revision

		1.	functor Krb_Tkt_File

 *)


functor Krb_Tkt_File(structure B : FOX_BASIS) : KRB_TKT_FILE =
struct

(*
	2.	types shared with Protocol
 *)
  type realm           = Protocol.realm
  type local_principal = Protocol.local_principal
  type credential      = Protocol.credential

(*
	3.	fun is_whitespace
 *)
  fun is_whitespace "" = true
    | is_whitespace " " = true
    | is_whitespace "\t" = true
    | is_whitespace "\n" = true
    | is_whitespace "\000" = true
    | is_whitespace _ = false

(*
	4.	fun input_string

	Read a string up to, but not including whitespace, newline, or eof.
 *)
  fun input_string f =
      case lookahead f of
        "" => ""
      | c  => if is_whitespace (input (f, 1))
                then ""
                else c ^ (input_string f)

(*
	5.	fun input_byte4

	Read a byte4 in host byte order.
 *)
  fun input_byte4 f =
      let
        fun c () = Byte1.from_int (ord (input (f,1)))
        val v = locative (Flat_Vector.flat_vector [c(), c(), c(), c()])
      in
        (Locative.locget (Locative.locview (v, desc_byte4))) : ubyte4
      end

(*
	6.	fun getenv
 *)
  fun getenv ename =
      let
        fun parse s =
          case B.V.String.index ("=", s, 0) of
            SOME k => (B.V.String.substring (s, 0, k),
                       B.V.String.substring (s, k+1, (size s)-(k+1)))
          | NONE => (s, "")
        fun parse_env [] = NONE
          | parse_env (first::rest) =
              case parse first of
                (name, value) => if name = ename
                                   then SOME value
                                   else parse_env rest
      in
        parse_env (System.Unsafe.CInterface.environ ())
      end

(*
	7.	fun ticketfile_name
 *)
  fun ticketfile_name () =
        case getenv "KRBTKFILE" of
          SOME filename => filename
        | NONE => "/tmp/tkt" ^ (Integer.makestring
                                 (System.Unsafe.CInterface.getpid ()))

(*
	8.	fun input_ticketfile

	Read the ticket file and parse it into a local_principal and
	list of credentials.
 *)
  fun input_ticketfile () =
      let
        val tf = open_in (ticketfile_name ())

        val byte4_to_byte1 = Byte1.from_int o Byte4.to_int

        fun list_credentials () =
          if end_of_stream tf then
            []
          else
            Protocol.Credential
              { service = Protocol.Principal
                            { name     = input_string tf,
                              instance = input_string tf,
                              realm    = input_string tf },
                session_key = Protocol.Key
                                (Convert.string2flat_vector (input (tf, 8))),
                lifetime = byte4_to_byte1 (input_byte4 tf),
                kvno     = byte4_to_byte1 (input_byte4 tf),
                ticket = let val len = input_byte4 tf in
                           Protocol.Ticket
                             (Convert.string2flat_vector
                                (input (tf, Byte4.to_int len)))
                         end,
                issue_date = input_byte4 tf}
            :: (list_credentials ())
      in
        ( { principal = Protocol.Local_Principal
                          { name     = input_string tf,
                            instance = input_string tf},
           credentials = list_credentials ()
          }
          before (close_in tf)
        )
        handle x => (close_in tf; raise x)
      end

(*
	9.	fun clear_ticketfile

	Clear the ticket file and write
 *)
  fun clear_ticketfile (Protocol.Local_Principal {name, instance}) =
    let
      val tf = open_out (ticketfile_name ())
    in
      output (tf, name     ^ "\000");
      output (tf, instance ^ "\000");
      close_out tf
    end

(*
	10.	fun append_ticketfile

	Add a credential to the end of the ticket file.
 *)
  fun append_ticketfile
        (Protocol.Credential
           { service as Protocol.Principal {name, instance, realm},
             session_key as Protocol.Key key,
             lifetime,
             kvno,
             ticket as Protocol.Ticket tkt,
             issue_date}) =
    let
      val tf = open_append (ticketfile_name ())

      val ticket_length = Byte4.from_int (Flat_Vector.flat_length tkt)

      val byte1_to_byte4 = Byte4.from_int o Byte1.to_int

      fun byte4_to_string x =
        let 
          val b = ByteArray.array (4,0)
          fun c k = chr (ByteArray.sub (b, k))
        in
          Byte4.update (b, 0, x);
          (c 0) ^ (c 1) ^ (c 2) ^ (c 3)
        end
    in
      output (tf, name     ^ "\000");
      output (tf, instance ^ "\000");
      output (tf, realm    ^ "\000");
      output (tf, Convert.flat_vector2string key);
      output (tf, byte4_to_string (byte1_to_byte4 lifetime));
      output (tf, byte4_to_string (byte1_to_byte4 kvno));
      output (tf, byte4_to_string (ticket_length));
      output (tf, Convert.flat_vector2string tkt);
      output (tf, byte4_to_string (issue_date));
      close_out tf
    end

end (* functor *)
